"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const TypeGraph_1 = require("./TypeGraph");
const Type_1 = require("./Type");
const Support_1 = require("./support/Support");
const Transformers_1 = require("./Transformers");
const TypeAttributes_1 = require("./TypeAttributes");
const StringTypes_1 = require("./StringTypes");
const Containers_1 = require("./support/Containers");
function transformationAttributes(graph, reconstitutedTargetType, transformer, debugPrintTransformation) {
    const transformation = new Transformers_1.Transformation(graph, reconstitutedTargetType, transformer);
    if (debugPrintTransformation) {
        console.log(`transformation for ${TypeGraph_1.typeRefIndex(reconstitutedTargetType)}:`);
        transformation.debugPrint();
    }
    return Transformers_1.transformationTypeAttributeKind.makeAttributes(transformation);
}
function makeEnumTransformer(graph, enumType, stringType, continuation) {
    const sortedCases = Array.from(enumType.cases).sort();
    const caseTransformers = sortedCases.map(c => new Transformers_1.StringMatchTransformer(graph, stringType, new Transformers_1.StringProducerTransformer(graph, stringType, continuation, c), c));
    return new Transformers_1.ChoiceTransformer(graph, stringType, caseTransformers);
}
function replaceUnion(union, builder, forwardingRef, debugPrintTransformations) {
    const graph = builder.typeGraph;
    Support_1.assert(union.members.size > 0, "We can't have empty unions");
    const reconstitutedMembersByKind = Containers_1.mapMapEntries(union.members.entries(), m => [
        m.kind,
        builder.reconstituteType(m)
    ]);
    const reconstitutedUnion = builder.getUnionType(union.getAttributes(), new Set(reconstitutedMembersByKind.values()));
    function memberForKind(kind) {
        return Support_1.defined(reconstitutedMembersByKind.get(kind));
    }
    function transformerForKind(kind) {
        const member = union.findMember(kind);
        if (member === undefined)
            return undefined;
        const memberTypeRef = Support_1.defined(reconstitutedMembersByKind.get(kind));
        return new Transformers_1.UnionInstantiationTransformer(graph, memberTypeRef);
    }
    let maybeStringType = undefined;
    function getStringType() {
        if (maybeStringType === undefined) {
            maybeStringType = builder.getStringType(TypeAttributes_1.emptyTypeAttributes, StringTypes_1.StringTypes.unrestricted);
        }
        return maybeStringType;
    }
    function transformerForStringType(t) {
        const memberRef = memberForKind(t.kind);
        switch (t.kind) {
            case "string":
                return Support_1.defined(transformerForKind(t.kind));
            case "date-time":
                return new Transformers_1.ParseDateTimeTransformer(graph, getStringType(), new Transformers_1.UnionInstantiationTransformer(graph, memberRef));
            case "enum": {
                const enumType = t;
                return makeEnumTransformer(graph, enumType, getStringType(), new Transformers_1.UnionInstantiationTransformer(graph, memberRef));
            }
            default:
                return Support_1.panic(`Can't transform string type ${t.kind}`);
        }
    }
    const stringTypes = union.stringTypeMembers;
    let transformerForString;
    if (stringTypes.size === 0) {
        transformerForString = undefined;
    }
    else if (stringTypes.size === 1) {
        const t = Support_1.defined(Containers_1.iterableFirst(stringTypes));
        transformerForString = new Transformers_1.UnionInstantiationTransformer(graph, memberForKind(t.kind));
    }
    else {
        transformerForString = new Transformers_1.ChoiceTransformer(graph, getStringType(), Array.from(stringTypes).map(transformerForStringType));
    }
    const transformerForClass = transformerForKind("class");
    const transformerForMap = transformerForKind("map");
    Support_1.assert(transformerForClass === undefined || transformerForMap === undefined, "Can't have both class and map in a transformed union");
    const transformerForObject = transformerForClass !== undefined ? transformerForClass : transformerForMap;
    const transformer = new Transformers_1.DecodingChoiceTransformer(graph, builder.getPrimitiveType("any"), transformerForKind("null"), transformerForKind("integer"), transformerForKind("double"), transformerForKind("bool"), transformerForString, transformerForKind("array"), transformerForObject);
    const attributes = transformationAttributes(graph, reconstitutedUnion, transformer, debugPrintTransformations);
    return builder.getPrimitiveType("any", attributes, forwardingRef);
}
function replaceEnum(enumType, builder, forwardingRef, debugPrintTransformations) {
    const stringType = builder.getStringType(TypeAttributes_1.emptyTypeAttributes, StringTypes_1.StringTypes.unrestricted);
    const transformer = new Transformers_1.DecodingTransformer(builder.typeGraph, stringType, makeEnumTransformer(builder.typeGraph, enumType, stringType));
    const reconstitutedEnum = builder.getEnumType(enumType.getAttributes(), enumType.cases);
    const attributes = transformationAttributes(builder.typeGraph, reconstitutedEnum, transformer, debugPrintTransformations);
    return builder.getStringType(attributes, StringTypes_1.StringTypes.unrestricted, forwardingRef);
}
function makeTransformations(graph, stringTypeMapping, targetLanguage, debugPrintTransformations, debugPrintReconstitution) {
    function replace(setOfOneUnion, builder, forwardingRef) {
        const t = Support_1.defined(Containers_1.iterableFirst(setOfOneUnion));
        if (t instanceof Type_1.UnionType) {
            return replaceUnion(t, builder, forwardingRef, debugPrintTransformations);
        }
        if (t instanceof Type_1.EnumType) {
            return replaceEnum(t, builder, forwardingRef, debugPrintTransformations);
        }
        return Support_1.panic(`Cannot make transformation for type ${t.kind}`);
    }
    const allTypesUnordered = graph.allTypesUnordered();
    const unions = Containers_1.setFilter(allTypesUnordered, t => t instanceof Type_1.UnionType && targetLanguage.needsTransformerForUnion(t));
    const enums = targetLanguage.needsTransformerForEnums
        ? Containers_1.setFilter(allTypesUnordered, t => t instanceof Type_1.EnumType)
        : new Set();
    const groups = Array.from(Containers_1.setUnion(unions, enums)).map(t => [t]);
    return graph.rewrite("make-transformations", stringTypeMapping, false, groups, debugPrintReconstitution, replace);
}
exports.makeTransformations = makeTransformations;
