"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const targetLanguages = require("./language/All");
const Support_1 = require("./support/Support");
const CombineClasses_1 = require("./rewrites/CombineClasses");
const InferMaps_1 = require("./rewrites/InferMaps");
const TypeBuilder_1 = require("./TypeBuilder");
const TypeGraph_1 = require("./TypeGraph");
const TypeNames_1 = require("./TypeNames");
const GatherNames_1 = require("./GatherNames");
const ExpandStrings_1 = require("./rewrites/ExpandStrings");
const FlattenUnions_1 = require("./rewrites/FlattenUnions");
const ResolveIntersections_1 = require("./rewrites/ResolveIntersections");
const ReplaceObjectType_1 = require("./rewrites/ReplaceObjectType");
const Messages_1 = require("./Messages");
const Inputs_1 = require("./input/Inputs");
const FlattenStrings_1 = require("./rewrites/FlattenStrings");
const MakeTransformations_1 = require("./MakeTransformations");
const Containers_1 = require("./support/Containers");
function getTargetLanguage(nameOrInstance) {
    if (typeof nameOrInstance === "object") {
        return nameOrInstance;
    }
    const language = targetLanguages.languageNamed(nameOrInstance);
    if (language !== undefined) {
        return language;
    }
    return Messages_1.messageError("DriverUnknownOutputLanguage", { lang: nameOrInstance });
}
exports.getTargetLanguage = getTargetLanguage;
const defaultOptions = {
    lang: "ts",
    inputData: new Inputs_1.InputData(),
    inferMaps: true,
    inferEnums: true,
    inferDates: true,
    alphabetizeProperties: false,
    allPropertiesOptional: false,
    combineClasses: true,
    fixedTopLevels: false,
    noRender: false,
    leadingComments: undefined,
    rendererOptions: {},
    indentation: undefined,
    outputFilename: "stdout",
    debugPrintGraph: false,
    checkProvenance: false,
    debugPrintReconstitution: false,
    debugPrintGatherNames: false,
    debugPrintTransformations: false,
    debugPrintTimes: false
};
class Run {
    constructor(options) {
        this._options = Object.assign(Object.assign({}, defaultOptions), options);
    }
    get stringTypeMapping() {
        const targetLanguage = getTargetLanguage(this._options.lang);
        return targetLanguage.stringTypeMapping;
    }
    timeSync(name, f) {
        return __awaiter(this, void 0, void 0, function* () {
            const start = Date.now();
            const result = yield f();
            const end = Date.now();
            if (this._options.debugPrintTimes) {
                console.log(`${name} took ${end - start}ms`);
            }
            return result;
        });
    }
    time(name, f) {
        const start = Date.now();
        const result = f();
        const end = Date.now();
        if (this._options.debugPrintTimes) {
            console.log(`${name} took ${end - start}ms`);
        }
        return result;
    }
    makeGraph(allInputs) {
        return __awaiter(this, void 0, void 0, function* () {
            const targetLanguage = getTargetLanguage(this._options.lang);
            const stringTypeMapping = targetLanguage.stringTypeMapping;
            const conflateNumbers = !targetLanguage.supportsUnionsWithBothNumberTypes;
            const typeBuilder = new TypeBuilder_1.TypeBuilder(0, stringTypeMapping, this._options.alphabetizeProperties, this._options.allPropertiesOptional, this._options.checkProvenance, false);
            yield this.timeSync("read input", () => __awaiter(this, void 0, void 0, function* () {
                return yield allInputs.addTypes(typeBuilder, this._options.inferMaps, this._options.inferEnums, this._options.inferDates, this._options.fixedTopLevels);
            }));
            let graph = typeBuilder.finish();
            if (this._options.debugPrintGraph) {
                graph.setPrintOnRewrite();
                graph.printGraph();
            }
            const debugPrintReconstitution = this._options.debugPrintReconstitution === true;
            if (typeBuilder.didAddForwardingIntersection) {
                this.time("remove indirection intersections", () => (graph = TypeGraph_1.removeIndirectionIntersections(graph, stringTypeMapping, debugPrintReconstitution)));
            }
            let unionsDone = false;
            if (allInputs.needSchemaProcessing) {
                let intersectionsDone = false;
                do {
                    const graphBeforeRewrites = graph;
                    if (!intersectionsDone) {
                        this.time("resolve intersections", () => ([graph, intersectionsDone] = ResolveIntersections_1.resolveIntersections(graph, stringTypeMapping, debugPrintReconstitution)));
                    }
                    if (!unionsDone) {
                        this.time("flatten unions", () => ([graph, unionsDone] = FlattenUnions_1.flattenUnions(graph, stringTypeMapping, conflateNumbers, true, debugPrintReconstitution)));
                    }
                    if (graph === graphBeforeRewrites) {
                        Support_1.assert(intersectionsDone && unionsDone, "Graph didn't change but we're not done");
                    }
                } while (!intersectionsDone || !unionsDone);
            }
            this.time("replace object type", () => (graph = ReplaceObjectType_1.replaceObjectType(graph, stringTypeMapping, conflateNumbers, targetLanguage.supportsFullObjectType, debugPrintReconstitution)));
            do {
                this.time("flatten unions", () => ([graph, unionsDone] = FlattenUnions_1.flattenUnions(graph, stringTypeMapping, conflateNumbers, false, debugPrintReconstitution)));
            } while (!unionsDone);
            if (this._options.combineClasses) {
                const combinedGraph = this.time("combine classes", () => CombineClasses_1.combineClasses(this, graph, this._options.alphabetizeProperties, true, false, debugPrintReconstitution));
                if (combinedGraph === graph) {
                    graph = combinedGraph;
                }
                else {
                    this.time("combine classes cleanup", () => (graph = CombineClasses_1.combineClasses(this, combinedGraph, this._options.alphabetizeProperties, false, true, debugPrintReconstitution)));
                }
            }
            if (this._options.inferMaps) {
                for (;;) {
                    const newGraph = this.time("infer maps", () => InferMaps_1.inferMaps(graph, stringTypeMapping, true, debugPrintReconstitution));
                    if (newGraph === graph) {
                        break;
                    }
                    graph = newGraph;
                }
            }
            const enumInference = allInputs.needSchemaProcessing ? "all" : this._options.inferEnums ? "infer" : "none";
            this.time("expand strings", () => (graph = ExpandStrings_1.expandStrings(graph, stringTypeMapping, enumInference, debugPrintReconstitution)));
            this.time("flatten unions", () => ([graph, unionsDone] = FlattenUnions_1.flattenUnions(graph, stringTypeMapping, conflateNumbers, false, debugPrintReconstitution)));
            Support_1.assert(unionsDone, "We should only have to flatten unions once after expanding strings");
            if (allInputs.needSchemaProcessing) {
                this.time("flatten strings", () => (graph = FlattenStrings_1.flattenStrings(graph, stringTypeMapping, debugPrintReconstitution)));
            }
            this.time("none to any", () => (graph = TypeGraph_1.noneToAny(graph, stringTypeMapping, debugPrintReconstitution)));
            if (!targetLanguage.supportsOptionalClassProperties) {
                this.time("optional to nullable", () => (graph = TypeGraph_1.optionalToNullable(graph, stringTypeMapping, debugPrintReconstitution)));
            }
            this.time("make transformations", () => (graph = MakeTransformations_1.makeTransformations(graph, stringTypeMapping, targetLanguage, this._options.debugPrintTransformations, debugPrintReconstitution)));
            this.time("flatten unions", () => ([graph, unionsDone] = FlattenUnions_1.flattenUnions(graph, stringTypeMapping, conflateNumbers, false, debugPrintReconstitution)));
            Support_1.assert(unionsDone, "We should only have to flatten unions once after making transformations");
            // Sometimes we combine classes in ways that will the order come out
            // differently compared to what it would be from the equivalent schema,
            // so we always just garbage collect to get a defined order and be done
            // with it.
            // FIXME: We don't actually have to do this if any of the above graph
            // rewrites did anything.  We could just check whether the current graph
            // is different from the one we started out with.
            this.time("GC", () => (graph = graph.garbageCollect(this._options.alphabetizeProperties, debugPrintReconstitution)));
            if (this._options.debugPrintGraph) {
                console.log("\n# gather names");
            }
            this.time("gather names", () => GatherNames_1.gatherNames(graph, this._options.debugPrintGatherNames));
            if (this._options.debugPrintGraph) {
                graph.printGraph();
            }
            return graph;
        });
    }
    makeSimpleTextResult(lines) {
        return new Map([[this._options.outputFilename, { lines, annotations: [] }]]);
    }
    run() {
        return __awaiter(this, void 0, void 0, function* () {
            // FIXME: This makes quicktype not quite reentrant
            TypeNames_1.initTypeNames();
            const targetLanguage = getTargetLanguage(this._options.lang);
            const inputData = this._options.inputData;
            yield inputData.finishAddingInputs();
            const needIR = inputData.needIR || targetLanguage.names.indexOf("schema") < 0;
            const schemaString = needIR ? undefined : inputData.singleStringSchemaSource();
            if (schemaString !== undefined) {
                const lines = JSON.stringify(JSON.parse(schemaString), undefined, 4).split("\n");
                lines.push("");
                const srr = { lines, annotations: [] };
                return new Map([[this._options.outputFilename, srr]]);
            }
            const graph = yield this.makeGraph(inputData);
            if (this._options.noRender) {
                return this.makeSimpleTextResult(["Done.", ""]);
            }
            return targetLanguage.renderGraphAndSerialize(graph, this._options.outputFilename, this._options.alphabetizeProperties, this._options.leadingComments, this._options.rendererOptions, this._options.indentation);
        });
    }
}
function quicktypeMultiFile(options) {
    return __awaiter(this, void 0, void 0, function* () {
        return yield new Run(options).run();
    });
}
exports.quicktypeMultiFile = quicktypeMultiFile;
function offsetLocation(loc, lineOffset) {
    return { line: loc.line + lineOffset, column: loc.column };
}
function offsetSpan(span, lineOffset) {
    return { start: offsetLocation(span.start, lineOffset), end: offsetLocation(span.end, lineOffset) };
}
function quicktype(options) {
    return __awaiter(this, void 0, void 0, function* () {
        const result = yield quicktypeMultiFile(options);
        if (result.size <= 1) {
            const first = Containers_1.mapFirst(result);
            if (first === undefined) {
                return { lines: [], annotations: [] };
            }
            return first;
        }
        let lines = [];
        let annotations = [];
        for (const [filename, srr] of result) {
            const offset = lines.length + 2;
            lines = lines.concat([`// ${filename}`, ""], srr.lines);
            annotations = annotations.concat(srr.annotations.map(ann => ({ annotation: ann.annotation, span: offsetSpan(ann.span, offset) })));
        }
        return { lines, annotations };
    });
}
exports.quicktype = quicktype;
