"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const TypeAttributes_1 = require("./TypeAttributes");
const Support_1 = require("./support/Support");
const Containers_1 = require("./support/Containers");
class StringTypes {
    // undefined means no restrictions
    constructor(cases, allowDate, allowTime, allowDateTime) {
        this.cases = cases;
        this.allowDate = allowDate;
        this.allowTime = allowTime;
        this.allowDateTime = allowDateTime;
        if (cases === undefined) {
            Support_1.assert(!this.allowDate && !this.allowTime && !this.allowDateTime, "We can't have an unrestricted string that also allows date/times");
        }
    }
    static fromCase(s, count) {
        const caseMap = {};
        caseMap[s] = count;
        return new StringTypes(new Map([[s, count]]), false, false, false);
    }
    static fromCases(cases) {
        const caseMap = {};
        for (const s of cases) {
            caseMap[s] = 1;
        }
        return new StringTypes(new Map(cases.map(s => [s, 1])), false, false, false);
    }
    get isRestricted() {
        return this.cases !== undefined;
    }
    union(othersArray, startIndex) {
        if (this.cases === undefined)
            return this;
        let cases = new Map(this.cases);
        let allowDate = this.allowDate;
        let allowTime = this.allowTime;
        let allowDateTime = this.allowDateTime;
        for (let i = startIndex; i < othersArray.length; i++) {
            const other = othersArray[i];
            if (other.cases === undefined)
                return other;
            Containers_1.mapMergeWithInto(cases, (x, y) => x + y, other.cases);
            allowDate = allowDate || other.allowDate;
            allowTime = allowTime || other.allowTime;
            allowDateTime = allowDateTime || other.allowDateTime;
        }
        return new StringTypes(cases, allowDate, allowTime, allowDateTime);
    }
    intersect(othersArray, startIndex) {
        let cases = this.cases;
        let allowDate = this.allowDate;
        let allowTime = this.allowTime;
        let allowDateTime = this.allowDateTime;
        for (let i = startIndex; i < othersArray.length; i++) {
            const other = othersArray[i];
            if (cases === undefined) {
                cases = Support_1.mapOptional(m => new Map(m), other.cases);
            }
            else if (other.cases !== undefined) {
                const thisCases = cases;
                const otherCases = other.cases;
                cases = Containers_1.mapMap(Containers_1.setIntersect(thisCases.keys(), new Set(otherCases.keys())).entries(), k => Math.min(Support_1.defined(thisCases.get(k)), Support_1.defined(otherCases.get(k))));
            }
            allowDate = allowDate && other.allowDate;
            allowTime = allowTime && other.allowTime;
            allowDateTime = allowDateTime && other.allowDateTime;
        }
        return new StringTypes(cases, allowDate, allowTime, allowDateTime);
    }
    applyStringTypeMapping(mapping) {
        if (!this.isRestricted)
            return this;
        const kinds = [];
        if (this.allowDate) {
            kinds.push(mapping.date);
        }
        if (this.allowTime) {
            kinds.push(mapping.time);
        }
        if (this.allowDateTime) {
            kinds.push(mapping.dateTime);
        }
        if (kinds.indexOf("string") >= 0) {
            return StringTypes.unrestricted;
        }
        const allowDate = kinds.indexOf("date") >= 0;
        const allowTime = kinds.indexOf("time") >= 0;
        const allowDateTime = kinds.indexOf("date-time") >= 0;
        return new StringTypes(this.cases, allowDate, allowTime, allowDateTime);
    }
    equals(other) {
        if (!(other instanceof StringTypes))
            return false;
        return (Containers_1.areEqual(this.cases, other.cases) &&
            this.allowDate === other.allowDate &&
            this.allowTime === other.allowTime &&
            this.allowDateTime === other.allowDateTime);
    }
    hashCode() {
        let h = Containers_1.hashCodeOf(this.cases);
        h = Support_1.addHashCode(h, Containers_1.hashCodeOf(this.allowDate));
        h = Support_1.addHashCode(h, Containers_1.hashCodeOf(this.allowTime));
        h = Support_1.addHashCode(h, Containers_1.hashCodeOf(this.allowDateTime));
        return h;
    }
    toString() {
        const parts = [];
        const enumCases = this.cases;
        if (enumCases === undefined) {
            parts.push("unrestricted");
        }
        else {
            const firstKey = Containers_1.iterableFirst(enumCases.keys());
            if (firstKey === undefined) {
                parts.push("enum with no cases");
            }
            else {
                parts.push(`${enumCases.size.toString()} enums: ${firstKey} (${enumCases.get(firstKey)}), ...`);
            }
        }
        if (this.allowDate)
            parts.push("d");
        if (this.allowTime)
            parts.push("t");
        if (this.allowDateTime)
            parts.push("dt");
        return parts.join(",");
    }
}
StringTypes.unrestricted = new StringTypes(undefined, false, false, false);
StringTypes.date = new StringTypes(new Map(), true, false, false);
StringTypes.time = new StringTypes(new Map(), false, true, false);
StringTypes.dateTime = new StringTypes(new Map(), false, false, true);
exports.StringTypes = StringTypes;
class StringTypesTypeAttributeKind extends TypeAttributes_1.TypeAttributeKind {
    constructor() {
        super("stringTypes");
    }
    get inIdentity() {
        return true;
    }
    requiresUniqueIdentity(st) {
        return st.cases !== undefined && st.cases.size > 0;
    }
    combine(arr) {
        Support_1.assert(arr.length > 0);
        return arr[0].union(arr, 1);
    }
    intersect(arr) {
        Support_1.assert(arr.length > 0);
        return arr[0].intersect(arr, 1);
    }
    makeInferred(_) {
        return undefined;
    }
    stringify(st) {
        return st.toString();
    }
}
exports.stringTypesTypeAttributeKind = new StringTypesTypeAttributeKind();
