"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const pluralize = require("pluralize");
const Support_1 = require("./support/Support");
const TypeAttributes_1 = require("./TypeAttributes");
const Strings_1 = require("./support/Strings");
const Chance_1 = require("./support/Chance");
const Containers_1 = require("./support/Containers");
let chance;
let usedRandomNames;
function initTypeNames() {
    chance = new Chance_1.Chance(31415);
    usedRandomNames = new Set();
}
exports.initTypeNames = initTypeNames;
initTypeNames();
function makeRandomName() {
    for (;;) {
        const name = `${chance.city()} ${chance.animal()}`;
        if (usedRandomNames.has(name))
            continue;
        usedRandomNames.add(name);
        return name;
    }
}
// FIXME: In the case of overlapping prefixes and suffixes we will
// produce a name that includes the overlap twice.  For example, for
// the names "aaa" and "aaaa" we have the common prefix "aaa" and the
// common suffix "aaa", so we will produce the combined name "aaaaaa".
function combineNames(names) {
    let originalFirst = Containers_1.iterableFirst(names);
    if (originalFirst === undefined) {
        return Support_1.panic("Named type has no names");
    }
    if (names.size === 1) {
        return originalFirst;
    }
    const namesSet = Containers_1.setMap(names, s => Strings_1.splitIntoWords(s)
        .map(w => w.word.toLowerCase())
        .join("_"));
    const first = Support_1.defined(Containers_1.iterableFirst(namesSet));
    if (namesSet.size === 1) {
        return first;
    }
    let prefixLength = first.length;
    let suffixLength = first.length;
    for (const n of Containers_1.iterableSkip(namesSet, 1)) {
        prefixLength = Math.min(prefixLength, n.length);
        for (let i = 0; i < prefixLength; i++) {
            if (first[i] !== n[i]) {
                prefixLength = i;
                break;
            }
        }
        suffixLength = Math.min(suffixLength, n.length);
        for (let i = 0; i < suffixLength; i++) {
            if (first[first.length - i - 1] !== n[n.length - i - 1]) {
                suffixLength = i;
                break;
            }
        }
    }
    const prefix = prefixLength > 2 ? first.substr(0, prefixLength) : "";
    const suffix = suffixLength > 2 ? first.substr(first.length - suffixLength) : "";
    const combined = prefix + suffix;
    if (combined.length > 2) {
        return combined;
    }
    return first;
}
exports.tooManyNamesThreshold = 20;
class TypeNames {
    static make(names, alternativeNames, areInferred) {
        if (names.size >= exports.tooManyNamesThreshold) {
            return new TooManyTypeNames(areInferred);
        }
        if (alternativeNames === undefined || alternativeNames.size > exports.tooManyNamesThreshold) {
            alternativeNames = undefined;
        }
        return new RegularTypeNames(names, alternativeNames, areInferred);
    }
}
exports.TypeNames = TypeNames;
class RegularTypeNames extends TypeNames {
    constructor(names, _alternativeNames, areInferred) {
        super();
        this.names = names;
        this._alternativeNames = _alternativeNames;
        this.areInferred = areInferred;
    }
    add(namesArray, startIndex = 0) {
        let newNames = new Set(this.names);
        let newAreInferred = this.areInferred;
        let newAlternativeNames = Support_1.mapOptional(s => new Set(s), this._alternativeNames);
        for (let i = startIndex; i < namesArray.length; i++) {
            const other = namesArray[i];
            if (!(other instanceof RegularTypeNames)) {
                Support_1.assert(other instanceof TooManyTypeNames, "Unknown TypeNames instance");
                if (!other.areInferred) {
                    // The other one is given, so it's the final word
                    return other;
                }
                if (newAreInferred === other.areInferred) {
                    // Both are inferred, so we let the TooMany sort it out
                    return other.add(namesArray, i + 1);
                }
                // Ours is given, the other inferred, so ignore it
                continue;
            }
            if (newAreInferred && !other.areInferred) {
                // Ours is inferred, the other one given, so take the
                // other's names
                newNames = new Set(other.names);
                newAreInferred = false;
            }
            else if (this.areInferred === other.areInferred) {
                Containers_1.setUnionInto(newNames, other.names);
            }
            if (other._alternativeNames !== undefined) {
                if (newAlternativeNames === undefined) {
                    newAlternativeNames = new Set();
                }
                Containers_1.setUnionInto(newAlternativeNames, other._alternativeNames);
            }
        }
        return TypeNames.make(newNames, newAlternativeNames, newAreInferred);
    }
    clearInferred() {
        const newNames = this.areInferred ? new Set() : this.names;
        return TypeNames.make(newNames, new Set(), this.areInferred);
    }
    get combinedName() {
        return combineNames(this.names);
    }
    get proposedNames() {
        const set = new Set([this.combinedName]);
        if (this._alternativeNames === undefined) {
            return set;
        }
        Containers_1.setUnionInto(set, this._alternativeNames);
        return set;
    }
    makeInferred() {
        if (this.areInferred)
            return this;
        return TypeNames.make(this.names, this._alternativeNames, true);
    }
    singularize() {
        return TypeNames.make(Containers_1.setMap(this.names, pluralize.singular), Support_1.mapOptional(an => Containers_1.setMap(an, pluralize.singular), this._alternativeNames), true);
    }
    toString() {
        const inferred = this.areInferred ? "inferred" : "given";
        const names = `${inferred} ${Array.from(this.names).join(",")}`;
        if (this._alternativeNames === undefined) {
            return names;
        }
        return `${names} (${Array.from(this._alternativeNames).join(",")})`;
    }
}
exports.RegularTypeNames = RegularTypeNames;
class TooManyTypeNames extends TypeNames {
    constructor(areInferred) {
        super();
        this.areInferred = areInferred;
        this.names = new Set([makeRandomName()]);
    }
    get combinedName() {
        return Support_1.defined(Containers_1.iterableFirst(this.names));
    }
    get proposedNames() {
        return this.names;
    }
    add(namesArray, startIndex = 0) {
        if (!this.areInferred)
            return this;
        for (let i = startIndex; i < namesArray.length; i++) {
            const other = namesArray[i];
            if (!other.areInferred) {
                return other.add(namesArray, i + 1);
            }
        }
        return this;
    }
    clearInferred() {
        if (!this.areInferred) {
            return this;
        }
        return TypeNames.make(new Set(), new Set(), true);
    }
    makeInferred() {
        return this;
    }
    singularize() {
        return this;
    }
    toString() {
        return `too many ${this.combinedName}`;
    }
}
exports.TooManyTypeNames = TooManyTypeNames;
class DescriptionTypeAttributeKind extends TypeAttributes_1.TypeAttributeKind {
    constructor() {
        super("names");
    }
    combine(namesArray) {
        Support_1.assert(namesArray.length > 0, "Can't combine zero type names");
        return namesArray[0].add(namesArray, 1);
    }
    makeInferred(tn) {
        return tn.makeInferred();
    }
    stringify(tn) {
        return tn.toString();
    }
}
exports.namesTypeAttributeKind = new DescriptionTypeAttributeKind();
function modifyTypeNames(attributes, modifier) {
    return exports.namesTypeAttributeKind.modifyInAttributes(attributes, modifier);
}
exports.modifyTypeNames = modifyTypeNames;
function singularizeTypeNames(attributes) {
    return modifyTypeNames(attributes, maybeNames => {
        if (maybeNames === undefined)
            return undefined;
        return maybeNames.singularize();
    });
}
exports.singularizeTypeNames = singularizeTypeNames;
function makeNamesTypeAttributes(nameOrNames, areNamesInferred) {
    let typeNames;
    if (typeof nameOrNames === "string") {
        typeNames = TypeNames.make(new Set([nameOrNames]), new Set(), Support_1.defined(areNamesInferred));
    }
    else {
        typeNames = nameOrNames;
    }
    return exports.namesTypeAttributeKind.makeAttributes(typeNames);
}
exports.makeNamesTypeAttributes = makeNamesTypeAttributes;
