"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const URI = require("urijs");
const JSONSchemaInput_1 = require("./JSONSchemaInput");
const CompressedJSON_1 = require("./CompressedJSON");
const JSONSchemaStore_1 = require("./JSONSchemaStore");
const Support_1 = require("../support/Support");
const Messages_1 = require("../Messages");
const TypeNames_1 = require("../TypeNames");
const Description_1 = require("../Description");
const Inference_1 = require("./Inference");
const All_1 = require("../language/All");
const AccessorNames_1 = require("../AccessorNames");
const Containers_1 = require("../support/Containers");
class InputJSONSchemaStore extends JSONSchemaStore_1.JSONSchemaStore {
    constructor(_inputs, _delegate) {
        super();
        this._inputs = _inputs;
        this._delegate = _delegate;
    }
    fetch(address) {
        return __awaiter(this, void 0, void 0, function* () {
            const maybeInput = this._inputs.get(address);
            if (maybeInput !== undefined) {
                return JSONSchemaInput_1.checkJSONSchema(Support_1.parseJSON(yield Support_1.toString(maybeInput), "JSON Schema", address), () => JSONSchemaInput_1.Ref.root(address));
            }
            if (this._delegate === undefined) {
                return Support_1.panic(`Schema URI ${address} requested, but no store given`);
            }
            return yield this._delegate.fetch(address);
        });
    }
}
class JSONInput {
    /* tslint:disable:no-unused-variable */
    constructor(_compressedJSON) {
        this._compressedJSON = _compressedJSON;
        this.kind = "json";
        this.needIR = true;
        this.needSchemaProcessing = false;
        this._topLevels = new Map();
    }
    addSample(topLevelName, sample) {
        let topLevel = this._topLevels.get(topLevelName);
        if (topLevel === undefined) {
            topLevel = { samples: [], description: undefined };
            this._topLevels.set(topLevelName, topLevel);
        }
        topLevel.samples.push(sample);
    }
    setDescription(topLevelName, description) {
        let topLevel = this._topLevels.get(topLevelName);
        if (topLevel === undefined) {
            return Support_1.panic("Trying to set description for a top-level that doesn't exist");
        }
        topLevel.description = description;
    }
    addSource(source) {
        return __awaiter(this, void 0, void 0, function* () {
            const { name, samples, description } = source;
            for (const sample of samples) {
                let value;
                try {
                    value = yield this._compressedJSON.readFromStream(Support_1.toReadable(sample));
                }
                catch (e) {
                    return Messages_1.messageError("MiscJSONParseError", {
                        description: Support_1.withDefault(description, "input"),
                        address: name,
                        message: Support_1.errorMessage(e)
                    });
                }
                this.addSample(name, value);
                if (description !== undefined) {
                    this.setDescription(name, description);
                }
            }
        });
    }
    finishAddingInputs() {
        return __awaiter(this, void 0, void 0, function* () {
            return;
        });
    }
    singleStringSchemaSource() {
        return undefined;
    }
    addTypes(typeBuilder, inferMaps, inferEnums, inferDates, fixedTopLevels) {
        return __awaiter(this, void 0, void 0, function* () {
            const inference = new Inference_1.TypeInference(this._compressedJSON, typeBuilder, inferMaps, inferEnums, inferDates);
            for (const [name, { samples, description }] of this._topLevels) {
                const tref = inference.inferType(TypeNames_1.makeNamesTypeAttributes(name, false), samples, fixedTopLevels);
                typeBuilder.addTopLevel(name, tref);
                if (description !== undefined) {
                    const attributes = Description_1.descriptionTypeAttributeKind.makeAttributes(new Set([description]));
                    typeBuilder.addAttributes(tref, attributes);
                }
            }
        });
    }
}
exports.JSONInput = JSONInput;
function jsonInputForTargetLanguage(targetLanguage, languages) {
    let lang;
    if (typeof targetLanguage === "string") {
        const maybeLang = All_1.languageNamed(targetLanguage, languages);
        if (maybeLang === undefined) {
            return Messages_1.messageError("DriverUnknownOutputLanguage", { lang: targetLanguage });
        }
        lang = maybeLang;
    }
    else {
        lang = targetLanguage;
    }
    const mapping = lang.stringTypeMapping;
    const makeDate = mapping.date !== "string";
    const makeTime = mapping.time !== "string";
    const makeDateTime = mapping.dateTime !== "string";
    const compressedJSON = new CompressedJSON_1.CompressedJSON(makeDate, makeTime, makeDateTime);
    return new JSONInput(compressedJSON);
}
exports.jsonInputForTargetLanguage = jsonInputForTargetLanguage;
class JSONSchemaInput {
    constructor(givenSchemaStore, additionalAttributeProducers = []) {
        this.kind = "schema";
        this.needSchemaProcessing = true;
        this._schemaStore = undefined;
        this._schemaInputs = new Map();
        this._schemaSources = [];
        this._topLevels = new Map();
        this._needIR = false;
        this._schemaStore = givenSchemaStore;
        this._attributeProducers = [Description_1.descriptionAttributeProducer, AccessorNames_1.accessorNamesAttributeProducer].concat(additionalAttributeProducers);
    }
    get needIR() {
        return this._needIR;
    }
    addTopLevel(name, ref) {
        this._topLevels.set(name, ref);
    }
    addTypes(typeBuilder) {
        return __awaiter(this, void 0, void 0, function* () {
            yield JSONSchemaInput_1.addTypesInSchema(typeBuilder, Support_1.defined(this._schemaStore), this._topLevels, this._attributeProducers);
        });
    }
    addSource(schemaSource) {
        return __awaiter(this, void 0, void 0, function* () {
            const { uris, schema, isConverted } = schemaSource;
            if (isConverted !== true) {
                this._needIR = true;
            }
            let normalizedURIs;
            const uriPath = `-${this._schemaInputs.size + 1}`;
            if (uris === undefined) {
                normalizedURIs = [new URI(uriPath)];
            }
            else {
                normalizedURIs = uris.map(uri => {
                    const normalizedURI = new URI(uri).normalize();
                    if (normalizedURI
                        .clone()
                        .hash("")
                        .toString() === "") {
                        normalizedURI.path(uriPath);
                    }
                    return normalizedURI;
                });
            }
            if (schema === undefined) {
                Support_1.assert(uris !== undefined, "URIs must be given if schema source is not specified");
            }
            else {
                for (const normalizedURI of normalizedURIs) {
                    this._schemaInputs.set(normalizedURI
                        .clone()
                        .hash("")
                        .toString(), schema);
                }
            }
            for (const normalizedURI of normalizedURIs) {
                this._schemaSources.push([normalizedURI, schemaSource]);
            }
        });
    }
    finishAddingInputs() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this._schemaSources.length === 0)
                return;
            let maybeSchemaStore = this._schemaStore;
            if (this._schemaInputs.size === 0) {
                if (maybeSchemaStore === undefined) {
                    return Support_1.panic("Must have a schema store to process JSON Schema");
                }
            }
            else {
                maybeSchemaStore = this._schemaStore = new InputJSONSchemaStore(this._schemaInputs, maybeSchemaStore);
            }
            const schemaStore = maybeSchemaStore;
            for (const [normalizedURI, source] of this._schemaSources) {
                const givenName = source.name;
                const refs = yield JSONSchemaInput_1.refsInSchemaForURI(schemaStore, normalizedURI, givenName);
                if (Array.isArray(refs)) {
                    let name;
                    if (this._schemaSources.length === 1) {
                        name = givenName;
                    }
                    else {
                        name = refs[0];
                    }
                    this.addTopLevel(name, refs[1]);
                }
                else {
                    for (const [refName, ref] of refs) {
                        this.addTopLevel(refName, ref);
                    }
                }
            }
        });
    }
    singleStringSchemaSource() {
        if (!this._schemaSources.every(([_, { schema }]) => typeof schema === "string")) {
            return undefined;
        }
        const set = new Set(this._schemaSources.map(([_, { schema }]) => schema));
        if (set.size === 1) {
            return Support_1.defined(Containers_1.iterableFirst(set));
        }
        return undefined;
    }
}
exports.JSONSchemaInput = JSONSchemaInput;
class InputData {
    constructor() {
        // FIXME: Make into a Map, indexed by kind.
        this._inputs = new Set();
    }
    addInput(input) {
        this._inputs = this._inputs.add(input);
    }
    addSource(kind, source, makeInput) {
        return __awaiter(this, void 0, void 0, function* () {
            let input = Containers_1.iterableFind(this._inputs, i => i.kind === kind);
            if (input === undefined) {
                input = makeInput();
                this.addInput(input);
            }
            yield input.addSource(source);
        });
    }
    finishAddingInputs() {
        return __awaiter(this, void 0, void 0, function* () {
            for (const input of this._inputs) {
                yield input.finishAddingInputs();
            }
        });
    }
    addTypes(typeBuilder, inferMaps, inferEnums, inferDates, fixedTopLevels) {
        return __awaiter(this, void 0, void 0, function* () {
            for (const input of this._inputs) {
                yield input.addTypes(typeBuilder, inferMaps, inferEnums, inferDates, fixedTopLevels);
            }
        });
    }
    get needIR() {
        return Containers_1.iterableSome(this._inputs, i => i.needIR);
    }
    get needSchemaProcessing() {
        return Containers_1.iterableSome(this._inputs, i => i.needSchemaProcessing);
    }
    singleStringSchemaSource() {
        const schemaStrings = Containers_1.setFilterMap(this._inputs, i => i.singleStringSchemaSource());
        if (schemaStrings.size > 1) {
            return Support_1.panic("We have more than one input with a string schema source");
        }
        return Containers_1.iterableFirst(schemaStrings);
    }
}
exports.InputData = InputData;
